﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public enum Facing {
	None,
	Right,
	Left
}

[RequireComponent(typeof(Rigidbody2D))]
public class Movement : MonoBehaviour {

	protected class Knockback {
		public string tag;

		public Vector2 velocity;
		public Vector2 acceleration;
		public float duration;

		public Knockback(Vector2 vel, Vector2 accel, float dur) {
			velocity = vel;
			acceleration = accel;
			duration = dur;
		}

		public Knockback(Vector2 vel, Vector2 accel, float dur, string newTag) {
			velocity = vel;
			acceleration = accel;
			duration = dur;

			tag = newTag;
		}
	}

	public Action groundedCallback;

	/**<summary>For testing purposes</summary>*/
	private bool bUseVelocityInstead = true;

	/**<summary>This gives the velocity from the last frame</summary>*/
	protected Vector2 velocityInternal;

	/** <summary> Returns a calculated velocity based on the amount moved since the last physics frame. </summary> */
	public Vector2 velocity {
		get {
			return velocityInternal;
		}
	}

	protected Vector2 currentMovementInternal;

	/** <summary> Set the speed used to try to move this object. </summary> */
	public Vector2 currentMovement {
		get {
			return currentMovementInternal;
		}
		set {
			currentMovementInternal = value;
			if (value.sqrMagnitude == 0) {
				bCurrentMovementIsZero = true;
			} else {
				bCurrentMovementIsZero = false;
			}
		}
	}
	protected bool bCurrentMovementIsZero = false;

	public bool IsGrounded {
		get {
			return bIsGrounded;
		}
	}
	protected bool bIsGrounded = false;

	private Facing _facing = Facing.Right;
	public Facing facing {
		get {
			return _facing;
		}
	}

	public Rigidbody2D Rigidbody2D { get { return rigidbody2D; } }

	new private Rigidbody2D rigidbody2D;

	private Vector2 lastPosition;

	private Vector2 _centerOfCollider;
	private Vector2 _rayExtension;
	private float _rayDistance;
	private int _terrainMask;

	private Vector2 _currentGravity;

	public bool bCanGravity = true;

	protected List<Knockback> knockbacks = new List<Knockback>();

	void Start() {
		rigidbody2D = GetComponent<Rigidbody2D>();
		rigidbody2D.interpolation = RigidbodyInterpolation2D.None;
		//animator = GetComponent<Animator>();

		Collider2D col2d = GetComponent<Collider2D>();
		_centerOfCollider = col2d.offset;
		_rayExtension = col2d.bounds.extents;
		_rayDistance = _rayExtension.y + 1f;
		_rayExtension.y = 1f;
		_rayExtension.x *= 2f;

		_terrainMask = 1 << 8;

		//Debug.Log(_centerOfCollider);
		//Debug.Log(_rayExtension);
		//Debug.Log(_rayDistance);

		lastPosition = transform.position;
	}

	void FixedUpdate() {
		velocityInternal = ((Vector2)transform.position - lastPosition) / Time.fixedDeltaTime;

		RaycastHit2D rayHit = Physics2D.BoxCast((Vector2)transform.position + _centerOfCollider, _rayExtension, 0, Vector2.down, _rayDistance, _terrainMask);

		if (rayHit.collider == null) {
			bIsGrounded = false;
		} else {
			if (!bIsGrounded) {
				bIsGrounded = true;
				if (groundedCallback != null) {
					groundedCallback();
				}
			}
		}

		if (bCanGravity) {
			if (!bIsGrounded) {
				_currentGravity += Physics2D.gravity * Time.fixedDeltaTime;

				if (_currentGravity.y < (Physics2D.gravity.y)) {
					_currentGravity = Physics2D.gravity;
				}
			} else {
				_currentGravity = Vector2.zero;
			}
		}

		// Distance traveled in a second
		Vector2 dist = currentMovementInternal;

		lastPosition = transform.position;

		int count = knockbacks.Count;
		for (int i = count - 1; i >= 0; i--) {
			if (knockbacks[i].duration > 0) {
				if (knockbacks[i].duration < Time.fixedDeltaTime) {
					dist += knockbacks[i].velocity; //* knockbacks[i].duration;
				} else {
					dist += knockbacks[i].velocity;
					knockbacks[i].velocity += knockbacks[i].acceleration;
				}
				knockbacks[i].duration -= Time.fixedDeltaTime;
			}
		}

		/*if (dist.sqrMagnitude >= 25 && rigidbody2D.collisionDetectionMode != CollisionDetectionMode2D.Continuous) {
			rigidbody2D.collisionDetectionMode = CollisionDetectionMode2D.Continuous;
		} else if (dist.sqrMagnitude < 25 && rigidbody2D.collisionDetectionMode != CollisionDetectionMode2D.Discrete) {
			rigidbody2D.collisionDetectionMode = CollisionDetectionMode2D.Discrete;
		}*/

		if (bUseVelocityInstead) {
			rigidbody2D.velocity = dist + _currentGravity;
		} else {
			if (dist != Vector2.zero) {
				rigidbody2D.MovePosition(lastPosition + (dist * Time.fixedDeltaTime) + _currentGravity);
			}
		}

		//Debug.Log(velocityInternal);
	}

	public void SetFacing(Facing newFacing) {
		_facing = newFacing;
	}

	public Vector2 FacingToVector() {
		switch (_facing) {
			//case Facing.Up:
				//return Vector2.up;
			case Facing.Right:
				return Vector2.right;
			//case Facing.Down:
				//return Vector2.down;
			case Facing.Left:
				return Vector2.left;
		}

		return Vector2.zero;
	}

	public void ReceiveKnockback(Vector2 vel, float duration) {
		ReceiveKnockback(vel, Vector2.zero, duration, null);
	}

	public void ReceiveKnockback(Vector2 vel, Vector2 accel, float duration) {
		ReceiveKnockback(vel, accel, duration, null);
	}

	public void ReceiveKnockback(Vector2 vel, Vector2 accel, float duration, string newTag) {
		int count = knockbacks.Count;
		for (int i = 0; i < count; i++) {
			if (knockbacks[i].duration <= 0) {
				knockbacks[i].velocity = vel;
				knockbacks[i].acceleration = accel;
				knockbacks[i].duration = duration;
				knockbacks[i].tag = newTag;

				return;
			}
		}

		knockbacks.Add(new Knockback(vel, accel, duration, newTag));
	}

	public void RemoveKnockback(string searchTag) {
		int count = knockbacks.Count;
		for (int i = 0; i < count; i++) {
			if (knockbacks[i].tag == searchTag) {
				knockbacks[i].velocity = Vector2.zero;
				knockbacks[i].acceleration = Vector2.zero;
				knockbacks[i].duration = -1;

				return;
			}
		}
	}

	public void SetGravity(bool bToggle) {
		bCanGravity = bToggle;
	}

	public void UseVelocityInstead(bool toggle) {
		bUseVelocityInstead = toggle;
	}

	public void Jump() {
		bIsGrounded = false;
	}
}
